
#include "cora_fit.h"
#include <stdio.h>

/* #define P_DBG */

int poly_fit (Vector * x, Vector * y, Vector * bgs, int m)
{
  int     i;

  double  a0 = 0.0, b0 = 0.0, c0 = 0.0;

  double  xt = 0.0, yt = 0.0;
  double  sx = 0.0, sy = 0.0, sxy = 0.0;
  double  det = 0.0, N = 0.0,  x0 = 0.0;
  double  sx2 = 0.0, sx2y = 0.0, sx3 = 0.0, sx4 = 0.0;

  double  tx = 0, ty = 0;

  /* trap inconsistencies and bomb out
   */
  if (x->len == 0 || x->vec == NULL)
    {
      fprintf (stderr, 
	       "* Warning *: poly_fit: No data in wavelength vector.\n");
      return (FALSE);
    }
  if (y->len == 0 || y->vec == NULL)
    {
      fprintf (stderr, 
	       "* Warning *: poly_fit: No data in flux vector.\n");
      return (FALSE);
    }
  if (bgs->vec == NULL)
    {
      fprintf (stderr, 
	       "* Warning *: poly_fit: Smoothed background not initialized.\n"
	       );
      return (FALSE);
    }
  if (m < 0 || m > 2)
    {
      fprintf (stderr, 
	       "* Warning *: poly_fit: Requested degree (%d) out of range"\
               " (allowed: 0 -- 2).\n", m);
      return (FALSE);
    }
  if (x->len != y->len || x->len != bgs->len)
    {
      fprintf (stderr, 
	       "* Warning *: poly_fit: Vectors have"\
	       " different lengths.\n");
      return (FALSE);
    }
    
  /* special case: one data point only 
   */
  if (x->len == 1)
    {
      bgs->vec[0] = y->vec[0];
      return (TRUE);
    }

  switch (m)
    {

      /* MITTELWERT 
       */
    case 0:
      for (i = 0; i < y->len; ++i)
	a0 += y->vec[i];
      a0 /= y->len;
      for (i = 0; i < y->len; ++i)
	bgs->vec[i] = a0;
      break;

      /* AUSGLEICHSGERADE
       */
    case 1:
      for (i = 0; i < y->len; ++i)
	{
	  yt = y->vec[i]; 
 	  xt = x->vec[i];
 
	  sx    += xt;
	  sy    += yt;
	  sxy   += xt * yt;

	  xt     = xt * xt;
	  sx2   += xt;
	}

      a0 = ( sy * sx2 - sxy * sx)   / (x->len * sx2 - sx * sx);
      b0 = (x->len * sxy - sx * sy) / (x->len * sx2 - sx * sx);

      for (i = 0; i < y->len; ++i)
	  bgs->vec[i] = a0 + b0 * x->vec[i];
      break;

      /* AUSGLEICHSPARABEL
       */
    case 2:

      /* transformation values
       */
      tx = (x->vec[0] + x->vec[y->len -1]) / 2;
      for (i = 0; i < y->len; ++i)
	ty += y->vec[i];
      ty /= y->len;

      for (i = 0; i < y->len; ++i)
	{
 	  x0 = x->vec[i] - tx;
	  yt = y->vec[i] - ty; 

#ifdef P_DBG
	  printf("x: %7.4f   y: %7.4f\n", x0, yt);
#endif

	  sx    += x0;
	  sy    += yt;
	  sxy   += (x0 * yt);

	  xt     = x0 * x0;
	  sx2   += xt;
	  sx2y  += (xt * yt);

	  xt     = xt * x0;
	  sx3   += xt;

	  xt     = xt * x0;
	  sx4   += xt;

	}

#ifdef P_DBG
      printf("sx:  %7.4f  sy:   %7.4f   sxy: %7.4f\n", 
	     sx, sy, sxy);
      printf("sx2: %7.4f  sx2y: %7.4f   sx3: %7.4f  sx4: %7.4f\n", 
	     sx2, sx2y, sx3, sx4);
#endif

      N   = (double) y->len;

      /**
      det =   N*sx2*sx4    + sx*sx3*sx2  + sx2*sx*sx3
            - sx2*sx2*sx2  - sx*sx*sx4   - N*sx3*sx3;

      a0  = sy*sx2*sx4     + sx*sx3*sx2y + sx2*sxy*sx3
            - sx2*sx2*sx2y - sy*sx3*sx3  - sx*sxy*sx4;

      b0  = N*sxy*sx4      + sy*sx3*sx2  + sx2*sx*sx2y
            - sx2*sxy*sx2  - sy*sx*sx4   - N*sx3*sx2y;

      c0  = N*sx2*sx2y     + sx*sxy*sx2  + sy*sx*sx3
            - N*sxy*sx3    - sx*sx*sx2y  - sy*sx2*sx2;
      **/

      det =   N*sx2*sx4  - sx2*sx2*sx2   + sx*sx3*sx2  
	- sx*sx*sx4 + sx2*sx*sx3  - N*sx3*sx3;

      a0  = sy*sx2*sx4     - sx2*sx2*sx2y + sx*sx3*sx2y
	- sy*sx3*sx3  + sx2*sxy*sx3 - sx*sxy*sx4;

      b0  = N*sxy*sx4      - sx2*sxy*sx2 + sy*sx3*sx2 
            - sy*sx*sx4   + sx2*sx*sx2y - N*sx3*sx2y;

      c0  = N*sx2*sx2y    - N*sxy*sx3  + sx*sxy*sx2 
            - sx*sx*sx2y  + sy*sx*sx3  - sy*sx2*sx2;


#ifdef P_DBG
      printf("DET: %7.4f  a0: %7.4f   b0: %7.4f  co: %7.4f\n", 
	     det, a0, b0, c0);
#endif

      a0 = a0 / det;
      b0 = b0 / det;
      c0 = c0 / det;


      /* 2 + x + x^2                   */
      /* a0 = 2.0; b0 = 1.0; c0 = 1.0; */

#ifdef P_DBG
      printf("DET: %7.4f  a0: %7.4f   b0: %7.4f  co: %7.4f\n", 
	     det, a0, b0, c0);
#endif

      for (i = 0; i < y->len; ++i)
	{
 	  xt = x->vec[i] - tx;
	  bgs->vec[i] = a0 + b0 * xt + c0 * xt * xt  + ty;
#ifdef P_DBG
	  printf("x: %7.4f   y: %7.4f  bg: %7.4f\n", 
		 x->vec[i], y->vec[i], bgs->vec[i]);
#endif

	}
      break;
    }
  return (TRUE);
}
